# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import copy
import numpy as np
cimport numpy as np
cimport cython

import pickle
import platform
from numpy import linalg as LA
from scipy import sparse
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
if platform.system() in ('Linux', 'Darwin'):
    from jax import device_put
    import jax.numpy as jnp


# column bind
def cbind(x, y, backend="cpu"):
    # if len(x.shape) == 1 or len(y.shape) == 1:
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        return jnp.column_stack((x, y))
    return np.column_stack((x, y))


# center... response
def center_response(y):
    if (len(np.asarray(y).shape)==1): 
        y_mean = np.mean(y)
        return y_mean, (y - y_mean)
    y_mean = np.asarray(y).mean(axis=0)    
    return y_mean, (y - y_mean[None, :])


# computes t(x)%*%y
def crossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = device_put(x)
        if y is None:
            return jnp.dot(x.T, x).block_until_ready()
        y = device_put(y)
        return jnp.dot(x.T, y).block_until_ready()
    if y is None:
        return np.dot(x.transpose(), x)
    return np.dot(x.transpose(), y)


# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# dropout
def dropout(x, drop_prob=0, seed=123):

    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)


# compute coeff in lasso
# Lasso via coordinate descent: the "Shooting Algorithm" of Fu (1998). Adapted from FDiTraglia's Github code +
# pseudocode algorithm 13.1 of Murphy (2012) + matlab code LassoShooting.m by Mark Schmidt.
def get_beta_1D(double[:] beta0, double[:,::1] XX2, 
                double[:] Xy2, double reg_lambda,
                int max_iter = 1000, 
                double tol = 1e-5):

    cdef int j, k, n_classes, p, converged, iteration
    cdef double aj, cj, err
    
    converged = 0
    err = 10000
    iteration = 0 
    p = len(beta0)
    beta_opt = np.asarray(beta0)

    while (converged != 1 and iteration < max_iter):
        beta_prev = pickle.loads(pickle.dumps(beta_opt, -1))
        for j in range(p):
            aj = XX2[j, j]
            cj = Xy2[j] - np.dot(np.asarray(XX2)[j, :], np.asarray(beta_opt)) + np.asarray(beta_opt)[j]*aj
            beta_opt[j] = soft_thres(cj/aj, reg_lambda/aj)
        err = np.sum(np.abs(np.asarray(beta_prev) - np.asarray(beta_opt))) 
        converged = (err <= tol)*1            
        iteration += 1  

    return np.asarray(beta_opt), iteration, err                    


# compute coeff in lasso
def get_beta_2D(double[:,::1] beta0, double[:,::1] XX2, 
                double[:,::1] Xy2, double reg_lambda,
                int max_iter = 1000, 
                double tol = 1e-5):

    cdef int j, k, n_classes, p, converged, iteration
    cdef double aj, cj, err
    cdef list ajs
    
    converged = 0
    err = 10000
    iteration = 0 
    p = len(beta0)
    beta_opt = np.asarray(beta0)

    # if len(beta0.shape) > 1: (multitask learner)
    n_classes = beta_opt.shape[1]
    ajs = [XX2[j, j] for j in range(p)]
    while (converged != 1 and iteration < max_iter):
        beta_prev = pickle.loads(pickle.dumps(beta_opt, -1))
        for k in range(n_classes):       
            beta_opt_k = np.asarray(beta_opt[:,k])                     
            for j in range(p):  
                aj = ajs[j]                         
                cj = Xy2[j, k] - np.dot(np.asarray(XX2)[j, :], beta_opt_k) + beta_opt_k[j]*aj
                beta_opt[j, k] = soft_thres(cj/aj, reg_lambda/aj)
        err = np.sqrt(np.sum(np.square(np.asarray(beta_prev) - np.asarray(beta_opt))))
        converged = (err <= tol)*1            
        iteration += 1  

    return np.asarray(beta_opt), iteration, err


# one-hot encoding
def one_hot_encode(x_clusters, n_clusters):

    assert (
        max(x_clusters) <= n_clusters
    ), "you must have max(x_clusters) <= n_clusters"

    n_obs = len(x_clusters)
    res = np.zeros((n_obs, n_clusters))

    for i in range(n_obs):
        res[i, x_clusters[i]] = 1

    return res


# one-hot encoding
def one_hot_encode2(y, n_classes):

    n_obs = len(y)
    res = np.zeros((n_obs, n_classes))

    for i in range(n_obs):
        res[i, y[i]] = 1

    return res


# row bind
def rbind(x, y, backend="cpu"):
    # if len(x.shape) == 1 or len(y.shape) == 1:
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        return jnp.row_stack((x, y))
    return np.row_stack((x, y))


# adapted from sklearn.utils.exmath
def safe_sparse_dot(a, b, backend="cpu", dense_output=False):
    """Dot product that handle the sparse matrix case correctly

    Parameters
    ----------
    a : array or sparse matrix
    b : array or sparse matrix
    dense_output : boolean, (default=False)
        When False, ``a`` and ``b`` both being sparse will yield sparse output.
        When True, output will always be a dense array.

    Returns
    -------
    dot_product : array or sparse matrix
        sparse if ``a`` and ``b`` are sparse and ``dense_output=False``.
    """
    sys_platform = platform.system()

    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        # modif when jax.scipy.sparse available
        return jnp.dot(device_put(a), device_put(b)).block_until_ready()

    #    if backend == "cpu":
    if a.ndim > 2 or b.ndim > 2:
        if sparse.issparse(a):
            # sparse is always 2D. Implies b is 3D+
            # [i, j] @ [k, ..., l, m, n] -> [i, k, ..., l, n]
            b_ = np.rollaxis(b, -2)
            b_2d = b_.reshape((b.shape[-2], -1))
            ret = a @ b_2d
            ret = ret.reshape(a.shape[0], *b_.shape[1:])
        elif sparse.issparse(b):
            # sparse is always 2D. Implies a is 3D+
            # [k, ..., l, m] @ [i, j] -> [k, ..., l, j]
            a_2d = a.reshape(-1, a.shape[-1])
            ret = a_2d @ b
            ret = ret.reshape(*a.shape[:-1], b.shape[1])
        else:
            ret = np.dot(a, b)
    else:
        ret = a @ b

    if (
        sparse.issparse(a)
        and sparse.issparse(b)
        and dense_output
        and hasattr(ret, "toarray")
    ):
        return ret.toarray()

    return ret

# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# scale... covariates
def scale_covariates(X, choice="std", training=True, scaler=None):

    scaling_options = {
        "std": StandardScaler(copy=True, with_mean=True, with_std=True),
        "minmax": MinMaxScaler(),
    }

    if training == True:
        # scaler must be not None
        scaler = scaling_options[choice]
        scaled_X = scaler.fit_transform(X)
        return scaler, scaled_X

    # training == False:
    # scaler must be not None
    return scaler.transform(X)

def soft_thres(double x, double y):
    return np.sign(x)*np.max(np.abs(x) - y, 0)

# from sklearn.utils.exmath
def squared_norm(x, backend="cpu"):
    """Squared Euclidean or Frobenius norm of x.

    Faster than norm(x) ** 2.

    Parameters
    ----------
    x : array_like

    Returns
    -------
    float
        The Euclidean norm when x is a vector, the Frobenius norm when x
        is a matrix (2-d array).
    """
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = np.ravel(x, order="K")
        x = device_put(x)        
        return jnp.dot(x, x).block_until_ready()

    x = np.ravel(x, order="K")
    return np.dot(x, x)


# computes x%*%t(y)
def tcrossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = device_put(x)
        if y is None:
            return jnp.dot(x, x.T).block_until_ready()
        y = device_put(y)
        return jnp.dot(x, y.T).block_until_ready()
    if y is None:
        return np.dot(x, x.transpose())
    return np.dot(x, y.transpose())


# convert vector to numpy array
def to_np_array(X):
    return np.array(X.copy(), ndmin=2)



# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: Thierry Moudiki
#
# License: BSD 3

import functools
import numpy as np
cimport numpy as np
cimport cython
import gc

from cython.parallel cimport prange
from libc.math cimport log, exp, sqrt, fabs
from numpy.linalg import lstsq
from numpy.linalg import norm
from scipy.special import expit
from sklearn.cluster import MiniBatchKMeans
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import StandardScaler
from multiprocessing import Pool
from tqdm import tqdm

from ..utils import subsample


# 0 - utils -----


# 0 - 0 data structures & funcs -----

# a tuple of doubles
cdef struct mydoubletuple:
    double elt1
    double elt2

ctypedef fused nparray_int:
    int
    
ctypedef fused nparray_double:    
    double
    
ctypedef fused nparray_long:
    long long

cdef dict __find_kmin_x_cache = {}
    


# 0 - 1 fitting -----

# returns max(x, y)
cdef double max_c(double x, double y):
    if (x > y):
        return x 
    return y


# returns min(x, y)
cdef double min_c(double x, double y):
    if (x < y):
        return x 
    return y


# sum vector 
cdef double cython_sum(double[:] x, long int n):
    cdef double res = 0 
    cdef long int i
    for i in range(n):
        res += x[i]
    return res


# finds index of a misclassed element 
cdef long int find_misclassed_index(nparray_long[:] misclass, 
                                    long int n_obs):    
    
    cdef long int i  
    for i in range(n_obs):        
        if (misclass[i] != 1000): # a badly classified example            
            return i    
    return 100000


# calculates sumproduct or 2 vectors
cdef double sum_product(long int[:] x, double[:] y, 
                        long int n):    
    cdef double res = 0
    cdef long int i    
    for i in range(n):
        if (x[i] != 0) & (y[i] != 0):
            res += x[i]*y[i]    
    return res


# calculates tolerance for early stopping
cdef double calculate_tolerance(nparray_double[:] x):     
    
    cdef long int n    
    cdef double ans 
    cdef nparray_double[:] temp
    
    if (len(x) >= 5): 
        
        x_ = np.asarray(x)                
        temp = x_[np.where(x_ != 0.0)] 
        n = len(temp)
        
        try:
            
            ans = (temp[n-1] - temp[(n-2)])        
            if (ans > 0):
                return ans
            return -ans
        
        except:
            
            return 1e5
        
    else:
        
        return 1e5
        
 
# calculates updating factor for misclassified example prob
cdef mydoubletuple calculate_diff_proba(long int n_obs, 
                         nparray_long[:] misclass,
                         nparray_double[:] w_new, 
                         nparray_double[:] w_prev,                          
                         double v_prev,
                         double eta,
                         double gamma):        
    
    cdef double v_new
    cdef double w_new_
    cdef double w_prev_
    cdef long int misclass_index
    
    misclass_index = find_misclassed_index(misclass, 
                                           n_obs)        
    
    if misclass_index != 1e5: 
    
        w_new_ = max_c(w_new[misclass_index], 
                       2.220446049250313e-16) 
            
        w_prev_ = w_prev[misclass_index]   
                
        v_new = (w_new_ - w_prev_)
                                                          
        return mydoubletuple(-(v_prev*gamma - v_new)/eta, v_new)                                                     
    
    return mydoubletuple(0, 0)
    

# updates probs
def update_proba(double[:,::1] probs, 
                 nparray_long[:] misclass,
                 nparray_double[:] w_new, 
                 nparray_double[:] w_prev,                 
                 double v_prev, double eta, double gamma, 
                 long int n_obs, int n_classes):        
    
    cdef double[:,::1] new_probs = np.asarray(probs) # contains result    
    cdef double diff_proba, new_p, prob    
    cdef long int idx    
    cdef int idy
    cdef double v_new
    
    
    res = calculate_diff_proba(n_obs, misclass, 
                               w_new, w_prev, v_prev, 
                               eta, gamma)     

    diff_proba = res.elt1
    v_new = res.elt2
                
    # loop on all examples 
    for idx in range(n_obs):
        
        if misclass[idx] != 1000: # a badly classified example                                                              
                        
            new_probs[idx, misclass[idx]] -= diff_proba # update proba
            # DO NOT normalize row (!)                         
    
    return tuple((new_probs, v_new))


# one-hot encoder for discrete response
def one_hot_encode(long int[:] y, 
                   int n_classes):
    
    cdef long int i 
    cdef long int n_obs = len(y)
    cdef double[:,::1] res = np.zeros((n_obs, n_classes))        

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)


# 0 - 2 predict -----

# squared norms
# @Paul Panzer's soln
# from https://stackoverflow.com/questions/42991347/how-to-find-the-pairwise-differences-between-rows-of-two-very-large-matrices-usi    
def outer_sum_dot(A, B):
    return np.add.outer((A*A).sum(axis=-1), (B*B).sum(axis=-1)) - 2*np.dot(A, B.T)
    

cdef double norm_c(double[:] x, long int n) nogil:
    
    cdef double res = 0
    
    for i in range(n):        
        res += x[i]**2
        
    return(sqrt(res))
    

cdef double euclidean_distance_c(double[:] x, 
                                 double[:] y,  
                                 long int n) nogil:
    
    cdef double res = 0
    
    for i in range(n):        
        res += (x[i] - y[i])**2
    
    return(sqrt(res))


cdef double cosine_similarity_c(double[:] x, 
                                double[:] y,  
                                long int n) nogil:
    
    cdef double dotprod = 0
    
    for i in range(n):        
        dotprod += (x[i]*y[i])**2
    
    return(1 - dotprod/(norm_c(x, n)*norm_c(y, n)))    
    

# distance of vector to matrix rows
# keep numpy arrays x, B    
cdef double manhattan_distance_c(double[:] x, 
                                 double[:] y,  
                                 long int n) nogil:
    
    cdef double res = 0
    
    for i in range(n):        
        res += fabs(x[i] - y[i])
    
    return(res)
    
    
# distance of vector to matrix rows
# keep numpy arrays x, B    
cdef nparray_double[:] distance_to_mat_euclidean(nparray_double[:] x, 
                   nparray_double[:,:] B):
    
    cdef long int i
    cdef long int n_B = B.shape[0]
    cdef long int p_B = B.shape[1]
    cdef nparray_double[:] res = np.zeros(n_B)
    

    for i in prange(n_B, nogil=True):

        res[i] = euclidean_distance_c(x, B[i, :], p_B)
                                
    return np.asarray(res)
    

# distance of vector to matrix rows
# keep numpy arrays x, B    
cdef nparray_double[:] distance_to_mat_manhattan(nparray_double[:] x, 
                   nparray_double[:,:] B):
    
    cdef long int i
    cdef long int n_B = B.shape[0]
    cdef long int p_B = B.shape[1]
    cdef nparray_double[:] res = np.zeros(n_B)
    

    for i in prange(n_B, nogil=True):

        res[i] = manhattan_distance_c(x, B[i, :], p_B)
                                
    return np.asarray(res)


# distance of vector to matrix rows
# keep numpy arrays x, B    
cdef nparray_double[:] distance_to_mat_cosine(nparray_double[:] x, 
                   nparray_double[:,:] B):
    
    cdef long int i
    cdef long int n_B = B.shape[0]
    cdef long int p_B = B.shape[1]
    cdef nparray_double[:] res = np.zeros(n_B)
    

    for i in prange(n_B, nogil=True):

        res[i] = cosine_similarity_c(x, B[i, :], p_B)        
                            
    return np.asarray(res)


# distance of vector to matrix rows
# keep numpy arrays x, B    
def distance_to_mat_euclidean2(nparray_double[:] x, 
                   nparray_double[:,:] B, double[:] res, 
                   long int n_B, long int p_B):
    
    cdef long int i

    for i in range(n_B):

        res[i] = euclidean_distance_c(x, B[i, :], p_B)
                            
    return res


# distance of vector to matrix rows
# keep numpy arrays x, B    
def distance_to_mat_manhattan2(nparray_double[:] x, 
                   nparray_double[:,:] B, double[:] res, 
                   long int n_B, long int p_B):
    
    cdef long int i

    for i in range(n_B):

        res[i] = manhattan_distance_c(x, B[i, :], p_B)
                            
    return res
        

# distance of vector to matrix rows
# keep numpy arrays x, B    
def distance_to_mat_cosine2(nparray_double[:] x, 
                   nparray_double[:,:] B, double[:] res, 
                   long int n_B, long int p_B):
    
    cdef long int i

    for i in range(n_B):

        res[i] = cosine_similarity_c(x, B[i, :], p_B)
                            
    return res

    
# find elt in list 
cdef long int find_elt_list(double elt, 
                            double[:] x, 
                            long int n_x):
    
    cdef long int j = 0
    
    for j in range(n_x):
        
        if x[j] == elt:
            
            return j        


# find_kmin_x
def find_kmin_x(x, n_x, k, cache=False):                 
    
    cdef str key 
    cdef double[:] sorted_x
    cdef dict res = {}
    cdef double elt = 0.0
    cdef int i = 0
    cdef long int idx 
    global __find_kmin_x_cache                          
    
    if cache: 
        
        key = str(hash(np.asarray(x).tostring())) 
        key += str(n_x) + str(k) 
    
        if key in __find_kmin_x_cache:
            
            return __find_kmin_x_cache[key]
        
        sorted_x = np.sort(x, kind='quicksort')        
    
        for i in range(k):
            
            elt = sorted_x[i]
            
            idx = find_elt_list(elt, x, n_x)
            
            res[idx] = elt
        
        __find_kmin_x_cache[key] = (np.asarray(list(res.values())), 
                                    np.asarray(list(res.keys()), 
                                    dtype=np.integer))                            
    
    # if cache == False    
    sorted_x = np.sort(x, kind='quicksort')        
    
    for i in range(k):
        
        elt = sorted_x[i]
        
        idx = find_elt_list(elt, x, n_x)
        
        res[idx] = elt
    
    return (np.asarray(list(res.values())), 
            np.asarray(list(res.keys()), 
            dtype=np.integer))                    


# calculate probs test i
def calculate_probs(long int[:] index, 
                    double[:,:] probs_train):
    
    cdef int idx, j
    cdef long int elt 
    cdef int n_classes = probs_train.shape[1]
    cdef double[:, ::1] probs_out = np.zeros((len(index), n_classes), 
                                              dtype=np.double)    
    
    idx = 0
    
    for elt in index:
        
        for j in range(n_classes):
        
            probs_out[idx, j] = probs_train[elt, j]
        
        idx += 1
    
    return np.asarray(probs_out)


# average probs
def average_probs(double[:,:] probs, double[:] weights):
    
    cdef double sum_probs
    cdef long int i 
    cdef long int n_obs = probs.shape[0]
    cdef int n_classes = probs.shape[1]
    #cdef double[:] probs_out = np.zeros(n_classes, dtype=np.double)
    cdef double[:] probs_out = np.zeros(n_classes)
    
    for j in range(n_classes):
        sum_probs = 0
        for i in range(n_obs):        
            sum_probs += probs[i, j]*weights[i]
        probs_out[j] = sum_probs
    
    return np.asarray(probs_out)
    

# calculate weights 
def calculate_weights(double[:] weights):
    cdef double[:] weights_out = np.zeros(len(weights))
    weights_out = 1/np.maximum(weights, np.finfo(float).eps)
    weights_out /= np.sum(weights_out)
    return weights_out


# 1 - model fitting -----

def fit_adaopt(double[:, :] X, long int[:] y, 
        int n_iterations,
        long int n_X, int p_X, 
        int n_classes,
        double learning_rate, 
        double reg_lambda, 
        double reg_alpha,
        double eta, 
        double gamma, 
        double tolerance):
    
    
    cdef double[:, :] Y
    cdef double[:, :] probs
    cdef double[:,:] scaled_X
    cdef double[:,:] beta
    cdef long int[:] preds    
    cdef double[:] w_prev
    cdef double[:] w_new
    cdef double[:] w_m
    cdef long int[:] misclass
    cdef long int[:] misclass_condition    
    cdef double err_m, alpha_m    
    cdef int m = 0    
    cdef double v_prev = 0 
    cdef double tolerance_   
    cdef double[:] alphas = np.zeros(n_iterations)   
    cdef double err_bound = (1.0-1.0/n_classes)
    
    
    # obtain initial probs for each example
    scaled_X = X/norm(X, ord=2, axis=1)[:, None]                        
    Y = one_hot_encode(y, n_classes) 
    beta = lstsq(scaled_X, Y, rcond=None)[0]
    probs = expit(np.dot(scaled_X, beta))
    probs /= np.sum(probs, axis=1)[:, None] 
    preds = np.asarray(probs).argmax(axis=1)    
    
    w_prev = np.repeat(1.0/n_X, n_X)    
    w_new = np.repeat(0.0, n_X)    
    w_m = np.repeat(1.0/n_X, n_X) 
    misclass = np.repeat(1000, n_X) 
    misclass_condition = np.repeat(0, n_X)  
    
    
    # loop update probs -----
    for m in range(n_iterations):
                
        # misclassification indicator for this iteration -----        
        for i in range(n_X):          
            if y[i] != preds[i]:                
                misclass[i] = preds[i]                
                misclass_condition[i] = 1  
            else:
                misclass[i] = 1000
                misclass_condition[i] = 0 
        
        
        # calculate error -----
        err_m = sum_product(misclass_condition, w_m, n_X) # sum(w_m) == 1  
        
        err_m += reg_lambda*(reg_alpha*cython_sum(np.abs(w_m), n_X) +\
                             (1 - reg_alpha)*0.5*cython_sum(np.power(w_m, 2), n_X))            
        err_m = min_c(max_c(err_m, 2.220446049250313e-16), 
                      err_bound)                
        
        
        # update weights based on misclassification error -----   
        alpha_m = learning_rate*log((n_classes - 1)*(1/err_m - 1))   
        alphas[m] = alpha_m
        w_prev = w_m     
        
        w_m *= np.exp(alpha_m*np.asarray(misclass_condition))
        w_m = np.asarray(w_m)/cython_sum(w_m, n_X)                                                    
        w_new = w_m
        
        # update probs based on misclassification error -----
        probs, v_prev = update_proba(probs, misclass,
                            w_new, w_prev, 
                            v_prev, eta, gamma, 
                            n_X, n_classes)        
        
        
        # recalculate preds -----
        preds = np.argmax(probs, axis=1) 
        
        
        # check tolerance -----               
        if  calculate_tolerance(np.asarray(alphas)) <= tolerance:
            n_iterations = m
            break               
    
    alphas_ = np.asarray(alphas)
    alphas_ = alphas_[np.where(alphas_ != 0)]
    probs_ = expit(np.asarray(probs))    
    probs_ /= np.sum(probs_, axis=1)[:, None]
        
    return {'probs': probs_, 
            'training_accuracy': np.mean(preds == np.asarray(y)),
            'alphas': alphas_, 
            'n_iterations': n_iterations, 
            'scaled_X_train': scaled_X}


# 2 - Model predict -----

def predict_proba_adaopt(double[:,::1] X_test, 
                  double[:,::1] scaled_X_train,
                  long int n_test, long int n_train,
                  double[:,::1] probs_train, int k,
                  int n_clusters, int seed,                   
                  int batch_size = 100,
                  type_dist="euclidean",    
                  cache=True):
    
    cdef int n_classes = probs_train.shape[1]        
    cdef int n_X_train = scaled_X_train.shape[0]
    cdef int p_X_train = scaled_X_train.shape[1]
    cdef double zero = np.finfo(float).eps
    cdef long int i, j
    cdef int m
    cdef tuple kmin_test_i
    cdef double[:] weights_test_i
    cdef long int[:] min_index_i, index_train
    
    cdef double[:,::1] dist_mat    
    cdef double[:,::1] out_probs = np.zeros((n_test, n_classes), 
                                            dtype=np.double)
    cdef double[:,::1] out_probs_ = np.zeros((n_test, n_classes), 
                                               dtype=np.double)
    cdef double[:,::1] scaled_X_test = X_test/norm(X_test, ord=2, axis=1)[:, None]
    cdef double[:,::1] probs_test_i = np.zeros((k, n_classes), 
                                               dtype=np.double)
    cdef double[:,::1] probs_train_ = np.zeros((n_train, n_classes), 
                                            dtype=np.double)
    cdef double[:] dists_test_i = np.zeros(n_train, dtype=np.double)
    cdef double[:] avg_probs_i = np.zeros(n_classes, dtype=np.double)
    cdef double[:] avg_probs = np.zeros(n_classes, dtype=np.double)
    
    
    # probabilities on training set -----
    
    if n_clusters <= 0: 

        # whole training set 
        probs_train_ = probs_train

    else: 

        # clustered training set 
        probs_train_ = np.zeros((n_clusters, n_classes), 
                                 dtype=np.double)

        kmeans = MiniBatchKMeans(n_clusters=n_clusters,
                                 batch_size=batch_size, 
                                 random_state=seed).fit(scaled_X_train)

        scaled_X_train = kmeans.cluster_centers_                        

        for m in range(n_clusters):

            index_train = np.where(kmeans.labels_ == m)[0]

            avg_probs = np.average(calculate_probs(index_train, 
                                                   probs_train), 
                                   axis=0)

            for j in range(n_classes):

                probs_train_[m, j] = avg_probs[j]            


    # main loops -----

    if type_dist == "euclidean":

        for i in range(n_test):        

            dists_test_i = distance_to_mat_euclidean(scaled_X_test[i,:], 
                                           scaled_X_train)        

            kmin_test_i = find_kmin_x(dists_test_i, 
                                      n_x=n_train, 
                                      k=k, cache=cache) 

            weights_test_i = calculate_weights(kmin_test_i[0])

            probs_test_i = calculate_probs(kmin_test_i[1], 
                                           probs_train_)  

            avg_probs_i = average_probs(probs=probs_test_i, 
                                        weights=weights_test_i)

            for j in range(n_classes):

                out_probs[i, j] = avg_probs_i[j]

        out_probs_ = expit(out_probs)  

        out_probs_ /= np.sum(out_probs_, axis=1)[:, None]

        return np.asarray(out_probs_)


    if type_dist == "cosine":

        for i in range(n_test):        

            dists_test_i = distance_to_mat_cosine(scaled_X_test[i,:], 
                                           scaled_X_train)        

            kmin_test_i = find_kmin_x(dists_test_i, 
                                      n_x=n_train, 
                                      k=k, cache=cache) 

            weights_test_i = calculate_weights(kmin_test_i[0])

            probs_test_i = calculate_probs(kmin_test_i[1], 
                                           probs_train_)  

            avg_probs_i = average_probs(probs=probs_test_i, 
                                        weights=weights_test_i)

            for j in range(n_classes):

                out_probs[i, j] = avg_probs_i[j]

        out_probs_ = expit(out_probs)  

        out_probs_ /= np.sum(out_probs_, axis=1)[:, None]

        return np.asarray(out_probs_)


    if type_dist == "manhattan":

        for i in range(n_test):        

            dists_test_i = distance_to_mat_manhattan(scaled_X_test[i,:], 
                                           scaled_X_train)        

            kmin_test_i = find_kmin_x(dists_test_i, 
                                      n_x=n_train, 
                                      k=k, cache=cache) 

            weights_test_i = calculate_weights(kmin_test_i[0])

            probs_test_i = calculate_probs(kmin_test_i[1], 
                                           probs_train_)  

            avg_probs_i = average_probs(probs=probs_test_i, 
                                        weights=weights_test_i)

            for j in range(n_classes):

                out_probs[i, j] = avg_probs_i[j]

        out_probs_ = expit(out_probs)  

        out_probs_ /= np.sum(out_probs_, axis=1)[:, None]

        return np.asarray(out_probs_)


    if type_dist == "euclidean-f":                

        dist_mat = outer_sum_dot(np.asarray(scaled_X_test), 
                                 np.asarray(scaled_X_train))        

        for i in range(n_test):        

            dists_test_i = dist_mat[i, :]

            kmin_test_i = find_kmin_x(dists_test_i, 
                                      n_x=n_train, 
                                      k=k, cache=cache)    

            weights_test_i = calculate_weights(kmin_test_i[0])

            probs_test_i = calculate_probs(kmin_test_i[1], 
                                           probs_train_) 

            avg_probs_i = average_probs(probs=probs_test_i, 
                                        weights=weights_test_i)

            for j in range(n_classes):

                out_probs[i, j] = avg_probs_i[j]

        out_probs_ = expit(out_probs)    
        out_probs_ /= np.sum(out_probs_, axis=1)[:, None]

        del dist_mat
        gc.collect()

        return np.asarray(out_probs_)    
   
    
def predict_adaopt(double[:,::1] X_test, 
            double[:,::1] scaled_X_train,
            long int n_test, long int n_train,
            double[:,::1] probs_train, int k,
            int n_clusters, int seed,                   
            int batch_size = 100,
            type_dist="euclidean",
            cache=True):            
    
    return np.asarray(predict_proba_adaopt(X_test, scaled_X_train,
                                    n_test, n_train,
                                    probs_train, k, 
                                    n_clusters, seed,
                                    batch_size, type_dist,
                                    cache)).argmax(axis=1)
    


import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from . import _stumpc as stumpc


class StumpClassifier(BaseEstimator, ClassifierMixin):
    """Stump classifier.
        
    Attributes:
    
        bins: int
            Number of histogram bins; as in numpy.histogram.        
    """

    def __init__(self, bins="auto"):

        self.bins = bins
        self.obj = None

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit Stump to training data (X, y)
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
        
            y: array-like, shape = [n_samples]
                Target values.
               
            sample_weight: array_like, shape = [n_samples]
                Observations weights.                                    
                      
        Returns:
        
            self: object.
        """

        if sample_weight is None:

            self.obj = stumpc.fit_stump_classifier(
                X=np.asarray(X, order="C"),
                y=np.asarray(y, order="C"),
                bins=self.bins,
            )

            return self

        self.obj = stumpc.fit_stump_classifier(
            X=np.asarray(X, order="C"),
            y=np.asarray(y, order="C"),
            sample_weight=np.ravel(sample_weight, order="C"),
            bins=self.bins,
        )

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
        
            **kwargs: additional parameters to be passed to `predict_proba`
                
               
        Returns: 
        
            model predictions: {array-like}
        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
        
            **kwargs: additional parameters to be passed to 
                self.cook_test_set
               
        Returns: 
        
            probability estimates for test data: {array-like}        
        """

        return stumpc.predict_proba_stump_classifier(
            self.obj, np.asarray(X, order="C")
        )



# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import functools
import pickle
import numpy as np
cimport numpy as np
cimport cython
import gc

from ..lasso import LassoRegressor
from ..ridge import RidgeRegressor
from cython.parallel cimport prange
from libc.math cimport log, exp, sqrt
from numpy.linalg import lstsq
from numpy.linalg import norm
from scipy.special import expit
from tqdm import tqdm


# 0 - utils -----

# 0 - 1 data structures & funcs -----

# a tuple of doubles
cdef struct mydoubletuple:
    double elt1
    double elt2

DTYPE_double = np.double

DTYPE_int = np.int

ctypedef np.double_t DTYPE_double_t


# 0 - 2 data structures & funcs -----

# dropout
def dropout_func(x, drop_prob=0, seed=123):

    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)
    
# one-hot encoder for discrete response
def one_hot_encode(long int[:] y, 
                   int n_classes):
    
    cdef long int i 
    cdef long int n_obs = len(y)
    cdef double[:,::1] res = np.zeros((n_obs, n_classes), dtype=DTYPE_double)        

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)
 
# 0 - 3 activation functions ----- 

def relu_activation(x):
    return np.maximum(x, 0)

def relu6_activation(x):
    return np.minimum(np.maximum(x, 0), 6)

def sigmoid_activation(x):
    return 1/(1 + np.exp(-x))

def activation_choice(x):
  activation_options = {
                "relu": relu_activation,
                "relu6": relu6_activation,
                "tanh": np.tanh,
                "sigmoid": sigmoid_activation
            }
  return activation_options[x]  


# 1 - classifier ----- 

# 1 - 1 fit classifier ----- 

def fit_booster_classifier(double[:,::1] X, long int[:] y, 
                           int n_estimators=100, double learning_rate=0.1, 
                           int n_hidden_features=5, double reg_lambda=0.1, 
                           double row_sample=1, double col_sample=1,
                           double dropout=0, double tolerance=1e-6, 
                           int direct_link=1, int verbose=1,
                           int seed=123, str backend="cpu", 
                           str solver="ridge", 
                           str activation="relu"
                           ): 
  
  cdef long int n
  cdef int p, n_cols_h_i, n_cols_hh_i
  cdef int n_classes
  cdef Py_ssize_t iter
  cdef dict res
  cdef double ym
  cdef double[:] xm, xsd
  cdef double[:,::1] Y, X_, E, W_i, h_i, hh_i, hidden_layer_i, hhidden_layer_i
  cdef double current_error
  
  
  n = X.shape[0]
  p = X.shape[1]
  res = {}
  current_error = 1000000.0
  
  xm = np.asarray(X).mean(axis=0)
  xsd = np.asarray(X).std(axis=0)
  
  res['direct_link'] = direct_link
  res['xm'] = np.asarray(xm)
  res['xsd'] = np.asarray(xsd)
  res['n_estimators'] = n_estimators
  res['learning_rate'] = learning_rate
  res['W_i'] = {}
  res['fit_obj_i'] = {} 
  res['col_index_i'] = {}
  res['loss'] = []
  res['activation'] = activation
  
  X_ = (np.asarray(X) - xm[None, :])/xsd[None, :]
  n_classes = len(np.unique(y))
  res['n_classes'] = n_classes
  
  Y = one_hot_encode(y, n_classes)
  Ym = np.mean(Y, axis=0)
  res['Ym'] = Ym
  E = Y - Ym
  iterator = tqdm(range(n_estimators)) if verbose else range(n_estimators)

  solvers = {"ridge": RidgeRegressor(reg_lambda = reg_lambda, backend = backend),
             "lasso": LassoRegressor(reg_lambda = reg_lambda, backend = backend)}
  fit_obj = solvers[solver]

  for iter in iterator:
      
      np.random.seed(seed + iter*1000)
    
      iy = np.sort(np.random.choice(a=range(p), 
                                    size=np.int(p*col_sample), 
                                    replace=False), 
                   kind='quicksort')
      res['col_index_i'][iter] = iy                     
      X_iy = np.asarray(X_)[:, iy] # must be X_!
      W_i = np.random.rand(X_iy.shape[1], n_hidden_features)
      hhidden_layer_i = dropout_func(x=activation_choice(activation)(np.dot(X_iy, W_i)), 
                                     drop_prob=dropout, seed=seed)
      hh_i = np.hstack((X_iy, hhidden_layer_i)) if direct_link else hhidden_layer_i
      
      if row_sample < 1:
      
        ix = np.sort(np.random.choice(a=range(n), 
                                    size=np.int(n*row_sample), 
                                    replace=False), 
                     kind='quicksort')
        X_iy_ix = X_iy[ix,:]       
        hidden_layer_i = dropout_func(x=activation_choice(activation)(np.dot(X_iy_ix, W_i)), 
                                      drop_prob=dropout, seed=seed)
        h_i =  np.hstack((X_iy_ix, hidden_layer_i)) if direct_link else hidden_layer_i
        fit_obj.fit(X = np.asarray(h_i), y = np.asarray(E)[ix,:])
                                 
      else:
      
        fit_obj.fit(X = np.asarray(hh_i), y = np.asarray(E))
            
      E = E - learning_rate*np.asarray(fit_obj.predict(np.asarray(hh_i)))
      
      res['W_i'][iter] = np.asarray(W_i)
            
      res['fit_obj_i'][iter] = pickle.loads(pickle.dumps(fit_obj, -1))

      current_error = np.linalg.norm(E, ord='fro')

      res['loss'].append(current_error)
      
      try:
        if np.abs(np.flip(np.diff(res['loss'])))[0] <= tolerance:
          res['n_estimators'] = iter
          break
      except:
        pass
      
  return res
  
  
# 1 - 2 predict classifier ----- 

def predict_proba_booster_classifier(object obj, double[:,::1] X):

  cdef int iter, n_estimators, n_classes
  cdef double learning_rate
  cdef double[:,::1] preds_sum, out_probs
  cdef long int n_row_preds 
  

  n_classes = obj['n_classes']
  direct_link = obj['direct_link']
  n_estimators = obj['n_estimators']
  learning_rate = obj['learning_rate']
  activation = obj['activation']
  X_ = (X - obj['xm'][None, :])/obj['xsd'][None, :]
  n_row_preds = X.shape[0]
  preds_sum = np.zeros((n_row_preds, n_classes))
  out_probs = np.zeros((n_row_preds, n_classes))
  
  
  for iter in range(n_estimators):
  
    iy = obj['col_index_i'][iter]
    X_iy = X_[:, iy] # must be X_!
    W_i = obj['W_i'][iter]
    hh_i = np.hstack((X_iy, activation_choice(activation)(np.dot(X_iy, W_i)))) if direct_link else activation_choice(activation)(np.dot(X_iy, W_i))
    preds_sum = preds_sum + learning_rate*np.asarray(obj['fit_obj_i'][iter].predict(np.asarray(hh_i)))
  
  out_probs = expit(np.tile(obj['Ym'], n_row_preds).reshape(n_row_preds, n_classes) + np.asarray(preds_sum))
  
  out_probs = out_probs/np.sum(out_probs, axis=1)[:, None]

  return np.asarray(out_probs)
  
  
# 2 - regressor -----   

# 2 - 1 fit regressor -----
  
def fit_booster_regressor(double[:,::1] X, double[:] y, 
                           int n_estimators=100, double learning_rate=0.1, 
                           int n_hidden_features=5, double reg_lambda=0.1, 
                           double row_sample=1, double col_sample=1,
                           double dropout=0, double tolerance=1e-6, 
                           int direct_link=1, int verbose=1, 
                           int seed=123, str backend="cpu", 
                           str solver="ridge", str activation="relu"): 
  
  cdef long int n
  cdef int p, n_cols_h_i, n_cols_hh_i
  cdef int n_classes
  cdef Py_ssize_t iter
  cdef dict res
  cdef double ym
  cdef double[:] xm, xsd, e
  cdef double[:,::1] X_, W_i, h_i, hh_i, hidden_layer_i, hhidden_layer_i
  cdef double current_error
  
  
  n = X.shape[0]
  p = X.shape[1]
  res = {}
  current_error = 1000000.0
  
  xm = np.asarray(X).mean(axis=0)
  xsd = np.asarray(X).std(axis=0)
  
  res['direct_link'] = direct_link
  res['xm'] = np.asarray(xm)
  res['xsd'] = np.asarray(xsd)
  res['n_estimators'] = n_estimators
  res['learning_rate'] = learning_rate
  res['activation'] = activation
  res['W_i'] = {}
  res['fit_obj_i'] = {} 
  res['col_index_i'] = {}
  res['loss'] = []
  
  X_ = (np.asarray(X) - xm[None, :])/xsd[None, :]
  n_classes = len(np.unique(y))
  res['n_classes'] = n_classes
  
  ym = np.mean(y)
  res['ym'] = ym
  e = y - np.repeat(ym, n)
  iterator = tqdm(range(n_estimators)) if verbose else range(n_estimators)

  solvers = {"ridge": RidgeRegressor(reg_lambda = reg_lambda, backend = backend),
             "lasso": LassoRegressor(reg_lambda = reg_lambda, backend = backend)}
  fit_obj = solvers[solver]

  for iter in iterator:
      
      np.random.seed(seed + iter*1000)
    
      iy = np.sort(np.random.choice(a=range(p), 
                                    size=np.int(p*col_sample), 
                                    replace=False), 
                   kind='quicksort')
      res['col_index_i'][iter] = iy                     
      X_iy = np.asarray(X_)[:, iy] # must be X_!
      W_i = np.random.rand(X_iy.shape[1], n_hidden_features)
      hhidden_layer_i = dropout_func(x=activation_choice(activation)(np.dot(X_iy, W_i)), 
                                     drop_prob=dropout, seed=seed)
      hh_i = np.hstack((X_iy, hhidden_layer_i)) if direct_link else hhidden_layer_i
      
      if row_sample < 1:
      
        ix = np.sort(np.random.choice(a=range(n), 
                                    size=np.int(n*row_sample), 
                                    replace=False), 
                     kind='quicksort')
        X_iy_ix = X_iy[ix,:]       
        hidden_layer_i = dropout_func(x=activation_choice(activation)(np.dot(X_iy_ix, W_i)), 
                                      drop_prob=dropout, seed=seed)
        h_i =  np.hstack((X_iy_ix, hidden_layer_i)) if direct_link else hidden_layer_i
        n_cols_h_i = h_i.shape[1]
        fit_obj.fit(X = np.asarray(h_i), y = np.asarray(e)[ix,:])

      else:
      
        fit_obj.fit(X = np.asarray(hh_i), y = np.asarray(e))
            
      e = e - learning_rate*np.asarray(fit_obj.predict(np.asarray(hh_i)))

      res['W_i'][iter] = np.asarray(W_i)
      
      res['fit_obj_i'][iter] = pickle.loads(pickle.dumps(fit_obj, -1))

      current_error = np.linalg.norm(e)

      res['loss'].append(current_error)      

      try:              
        if np.abs(np.flip(np.diff(res['loss'])))[0] <= tolerance:
          res['n_estimators'] = iter
          break      
      except:
        pass
      
  return res
  
# 2 - 2 predict regressor -----

def predict_booster_regressor(object obj, double[:,::1] X):

  cdef int iter, n_estimators, n_classes
  cdef double learning_rate
  cdef double[:] preds_sum
  cdef double[:,::1] hh_i

  n_classes = obj['n_classes']
  direct_link = obj['direct_link']
  n_estimators = obj['n_estimators']
  learning_rate = obj['learning_rate']
  activation = obj['activation']
  X_ = (X - obj['xm'][None, :])/obj['xsd'][None, :]
  preds_sum = np.zeros(X.shape[0])
  
  for iter in range(n_estimators):
  
    iy = obj['col_index_i'][iter]
    X_iy = X_[:, iy] # must be X_!
    W_i = obj['W_i'][iter]
    hh_i = np.hstack((X_iy, activation_choice(activation)(np.dot(X_iy, W_i)))) if direct_link else activation_choice(activation)(np.dot(X_iy, W_i))    
    preds_sum = preds_sum + learning_rate*np.asarray(obj['fit_obj_i'][iter].predict(np.asarray(hh_i)))
  
  return np.asarray(obj['ym'] + np.asarray(preds_sum))


# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import functools
import numpy as np
cimport numpy as np
cimport cython
import gc
import operator
import warnings
import pickle

from collections import Counter
from cython.parallel cimport prange
from libc.math cimport log, exp, sqrt, fabs
from numpy.linalg import lstsq
from numpy.linalg import norm
from scipy.special import expit



# 0 - utils -----

warnings.filterwarnings("ignore")

# returns min(x, y)
cdef double min_c(double x, double y):
    if (x < y):
        return x 
    return y

# returns max(x, y)
cdef double max_c(double x, double y):
    if (x > y):
        return x 
    return y


# 0 - 1 data structures & funcs -----

# a tuple of doubles
cdef struct mydoubletuple:
    double elt1
    double elt2

ctypedef fused nparray_int:
    int
    
ctypedef fused nparray_double:    
    double
    
ctypedef fused nparray_long:
    long long


# 0 - 2 data structures & funcs -----

# one-hot encoder for discrete response
def one_hot_encode(long int[:] y, 
                   int n_classes):
    
    cdef long int i 
    cdef long int n_obs = len(y)
    cdef double[:,::1] res = np.zeros((n_obs, n_classes))        

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)

# average stuff
cdef double cython_average(double[:] x, double[:] weights):
    
    cdef long int i, n 
    cdef double sum_x    
    cdef double[:] sum_out 
    
    n = len(x)
    sum_x = 0 

    for i in range(n):        
      sum_x += x[i]*weights[i]        
    
    return sum_x

# obtain histogram bins        
def histogram_bins(double[:] x, long int n_bins = -1):
    
    cdef long int i, n    
    cdef long int n_classes, n_bins_1 
    cdef double freq     
    cdef double[:] res
    cdef double min_x, max_x

    n = len(x)    

    if n_bins == -1:
      n_bins = int(sqrt(n))    

    min_x = np.min(x)    
    max_x = np.max(x)

    freq = (max_x - min_x)/n_bins
    
    n_bins_1 = n_bins + 1

    res = np.zeros(n_bins_1)

    for i in range(n_bins_1):

      res[i] = min_x + i*freq
                                
    return np.asarray(res)

   
# 1 - fitting -----    


def fit_stump_classifier(double[:,::1] X, long int[:] y, 
                         double[:] sample_weight=None, 
                         bins="auto"):

  cdef long int n
  cdef long int n_up 
  cdef long int n_down 
  cdef long int n_preds
  cdef int n_bins 
  cdef int p
  cdef long int i
  cdef int j, k
  cdef int best_col
  cdef double[:] cutpoints
  cdef double cutpoint_i
  cdef double best_cutpoint
  cdef long int n_cutpoints
  cdef double error_rate, error_rate_cur, error_rate_down, error_rate_up
  cdef int best_class, class_up, n_classes
  cdef dict res_class
  cdef double[:] equal_weights


  n = X.shape[0]
  p = X.shape[1]
  n_classes = len(np.unique(y))
  res_class = {}
  error_rate = 10000.0
  error_rate_cur = 10000.0
  best_cutpoint = 0.0
  i = 0
  j = 0
  k = 0
  if bins == "auto":
    n_bins = -1
  else: 
    n_bins = bins  
  n_preds = len(y)  
  equal_weights = np.repeat(1/n_preds, n_preds)
  
  X_ = np.asarray(X).T.tolist()
  
  if n_classes <= 2:    
  
    for j in range(p):
    
      X_j = np.asarray(X_[j])
      
      cutpoints = histogram_bins(X_j, n_bins) 

      n_cutpoints = len(cutpoints)
      
      for i in range(n_cutpoints):
               
        cutpoint_i = cutpoints[i] 
        index_up = (X_j <= cutpoint_i)
        y_up = np.asarray(y)[index_up]        
        class_up = Counter(y_up.tolist()).most_common(1)[0][0] # majority vote # majority vote

        preds = class_up*index_up + (1 - class_up)*np.logical_not(index_up)                
        
        error_rate_cur = cython_average((preds != y)*1.0, equal_weights) if sample_weight is None else cython_average((preds != y)*1.0, sample_weight)           
        
        if error_rate_cur <= error_rate:
          # print(error_rate_cur)
          best_col = j
          best_cutpoint = cutpoint_i
          error_rate = error_rate_cur
          best_class = class_up
         
    return best_col, best_cutpoint, best_class

# if n_classes > 2:

  Y = one_hot_encode(y, n_classes)
  Y_ = Y.T.tolist()
  
  for k in range(n_classes):
  
    y_ = np.asarray(Y_[k])
    best_col = 0
    best_cutpoint = 0.0
    best_class = 0
    error_rate = 10000.0
    error_rate_cur = 10000.0
    best_cutpoint = 0.0
  
    for j in range(p):
    
      X_j = np.asarray(X_[j])
           
      cutpoints = histogram_bins(X_j, n_bins) 

      n_cutpoints = len(cutpoints)      
 
      for i in range(n_cutpoints):
        
        try:
        
          cutpoint_i = cutpoints[i] 
          index_up = (X_j <= cutpoint_i)
          y_up = y_[index_up]          
          class_up = Counter(y_up.tolist()).most_common(1)[0][0] # majority vote
  
          preds = class_up*index_up + (1 - class_up)*np.logical_not(index_up)
                     
          error_rate_cur = cython_average((preds != y_)*1.0, equal_weights) if sample_weight is None else cython_average((preds != y_)*1.0, sample_weight)                     
          
          if error_rate_cur <= error_rate:
            # print(error_rate_cur)
            best_col = j
            best_cutpoint = cutpoint_i
            error_rate = error_rate_cur
            best_class = class_up
  
        except:
        
          pass
    
    res_class[k] = (best_col, best_cutpoint, best_class)
        
  return res_class
  
  
  
# 2 - predict -----   

def predict_proba_twoclass(double[:,::1] X, 
                             int best_col, 
                             double best_cutpoint, 
                             int best_class):
    X_ = np.asarray(X).T.tolist()
    X_j = np.asarray(X_[best_col])
    preds = (X_j <= best_cutpoint)*best_class + (X_j > best_cutpoint)*(1 - best_class)
    return (np.asarray(one_hot_encode(preds, 2)))

def predict_proba_stump_classifier(object obj, double[:,::1] X):
  
  cdef long int n_obs
  cdef int n_classes
  cdef double[:,:] raw_probs
  cdef long int[:] temp_prob


  # if n_classes <= 2  
  if (isinstance(obj, dict) == False):
    return (predict_proba_twoclass(X, obj[0], obj[1], obj[2]))
    
    
  # else: n_classes > 2
  n_obs = X.shape[0]
  n_classes = len(obj)
  raw_probs = np.zeros((n_obs, n_classes))
  
  
  for k in range(n_classes):
    # print(f"class {k}")
    # print(predict_proba_twoclass(X, obj[k][0], obj[k][1], obj[k][2]))
    temp_prob = np.argmax(predict_proba_twoclass(X, obj[k][0], obj[k][1], obj[k][2]), 
                              axis=1)
    for i in range(n_obs):
      raw_probs[i, k] = temp_prob[i]
  
  # print("raw_probs")
  # print(np.asarray(raw_probs))
  expit_raw_probs = expit(raw_probs)
  
  return (expit_raw_probs/expit_raw_probs.sum(axis=1)[:, None])
  
  
  
    
  